// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test7mat.cpp
// Purpose:	Time Measurements for Rule Implementations: Materialization
// Last Change:	26.01.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Test Program 7:
//
// p(X, Z, 2) :- q(X, Y), r(Y, 5, Z).
// q(X, Y) :- s(X, Y), Y >= 0.
//
// r = {(i,j,k) | 1 <= i <= 10000, 1 <= j <= 100, 1 <= k <= 100}
//	r can be accessed with binding pattern bff (index on first argument)
// s = {(i,i) | 1 <= i <= 10000 }

#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int S_TUPLES = 10000;

const int R_COL1_VALUES =  10000;

const int R_COL2_VALUES = 100;

const int R_COL3_VALUES = 100;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Predicate p - implemented as linked list
// ----------------------------------------------------------------------------

class p {
	friend class p_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }

	static void insert(int c1, int c2, int c3) {
		p *new_elem = new p(c1, c2, c3);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p(int c1, int c2, int c3)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	p *Next;

	static class p* First;
	static class p* Last;
};

class p* p::First;
class p* p::Last;

class p_cursor {
public:
	p_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() { return Current->col_1(); }
	int col_2() { return Current->col_2(); }
	int col_3() { return Current->col_3(); }

	void close() {}

private:
	bool Before_First;
	p*   Current;
};

// ----------------------------------------------------------------------------
// Predicate q - implemented as linked list
// ----------------------------------------------------------------------------

class q {
	friend class q_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		q *new_elem = new q(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	q(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	q *Next;

	static class q* First;
	static class q* Last;
};

class q* q::First;
class q* q::Last;

class q_cursor {
public:
	q_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = q::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() { return Current->col_1(); }
	int col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	q*   Current;
};

//-----------------------------------------------------------------------------
// Predicate s: Tuples generated by computation
//-----------------------------------------------------------------------------

class s_cursor {
public:
	s_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < S_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i; }

	void close() {}

private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r (binding pattern bff): Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor_bff {
public:
	r_cursor_bff() { }

	void open(int c1)
	{
		i = c1;
		if(i > 0 && i <= R_COL1_VALUES) {
			j = 0;
			k = R_COL3_VALUES;
		}
		else {
			j = R_COL2_VALUES + 1;
			k = 0;
		}
	}

	bool fetch()
	{
		if(j <= R_COL2_VALUES && k < R_COL3_VALUES) {
			k++;
			return true;
		}
		else if(j < R_COL2_VALUES) {
			j++;
			k = 1;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return j; }

	int col_3() { return k; }

	void close() {}

private:
	int i;
	int j;
	int k;
};


//-----------------------------------------------------------------------------
// Implementation of Rule q(X, Y) :- s(X, Y), Y >= 0: Materialization
//-----------------------------------------------------------------------------

void rule1_mat()
{

	s_cursor s1;
	s1.open();
	while(s1.fetch()) {
		int X = s1.col_1();
		int Y = s1.col_2();
		if(Y >= 0)
			q::insert(X, Y);
	}
	s1.close();
}

//-----------------------------------------------------------------------------
// Implementation of Rule p(X, Z, 2) :- q(X, Y), r(Y, 5, Z): Materialization
//-----------------------------------------------------------------------------

void rule2_mat()
{

// Program code copied from paper:
q_cursor q1;
q1.open();
while(q1.fetch()) {
    int X = q1.col_1();
    int Y = q1.col_2();
    r_cursor_bff r1;
    r1.open(Y); // if there is an index on the first arg
    while(r1.fetch()) {
        if(r1.col_2() == 5) {
            int Z = r1.col_3();
            p::insert(X, Z, 2);
        }
    }
}

}

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;

	cout << "Test of Method 1: Materialization (Program 7)\n";
	t.start();
	rule1_mat();
	rule2_mat();
	p_cursor p1;
	p1.open();
	while(p1.fetch()) {
		c1 += p1.col_1();
		c2 += p1.col_2();
		c3 += p1.col_3();
	}
	p1.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p1.open();
		while(p1.fetch()) {
			cout << "p(" << p1.col_1() << ", " << p1.col_2()
				<< ", " << p1.col_3() << ").\n";
		}
		p1.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << ", col3="
		<< c3 << "\n\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n";

	return 0;
}


